use anyhow::{bail, Result};
use cgmath::{Deg, Point2};
use std::collections::HashMap;

use crate::physics::{Pfloat, Polar};

/// Toml file syntax
pub(in crate::content) mod toml {
	use super::Pfloat;
	use serde::Deserialize;
	use std::collections::HashMap;

	#[derive(Debug, Deserialize)]
	pub struct SystemRoot {
		pub system: System,
		pub object: HashMap<String, Object>,
	}

	#[derive(Debug, Deserialize)]
	pub struct System {
		pub name: String,
	}

	#[derive(Debug, Deserialize)]
	pub struct Object {
		pub sprite: String,
		pub center: Center,

		pub size: Pfloat,
		pub parallax: Pfloat,

		#[serde(default)]
		pub radius: Pfloat,
		#[serde(default)]
		pub angle: Pfloat,
	}

	#[derive(Debug, Deserialize)]
	#[serde(untagged)]
	pub enum Center {
		Label(String),
		Coords([Pfloat; 2]),
	}

	impl ToString for Center {
		fn to_string(&self) -> String {
			match self {
				Self::Label(s) => s.to_owned(),
				Self::Coords(v) => format!("{:?}", v),
			}
		}
	}
}

#[derive(Debug)]
pub struct System {
	pub name: String,
	pub objects: Vec<Object>,
}

#[derive(Debug)]
pub struct Object {
	pub sprite: String,
	pub position: Polar,
	pub size: Pfloat,
	pub parallax: Pfloat,
}

fn resolve_center(
	objects: &HashMap<String, toml::Object>,
	obj: &toml::Object,
) -> Option<Point2<f32>> {
	match &obj.center {
		toml::Center::Label(s) => Some(
			Polar {
				center: resolve_center(&objects, objects.get(s)?)?,
				radius: obj.radius,
				angle: Deg(obj.angle),
			}
			.to_cartesian(),
		),
		toml::Center::Coords(v) => Some((*v).into()),
	}
}

impl System {
	pub fn parse(value: toml::SystemRoot) -> Result<Self> {
		let mut objects = Vec::new();

		for (label, obj) in &value.object {
			let center = match resolve_center(&value.object, obj) {
				Some(v) => v,
				None => {
					bail!(
						"could not resolve center label `{}` in `object.{}`",
						obj.center.to_string(),
						label
					);
				}
			};

			objects.push(Object {
				sprite: obj.sprite.clone(),
				position: Polar {
					center,
					radius: obj.radius,
					angle: Deg(obj.angle),
				},
				size: obj.size,
				parallax: obj.parallax,
			});
		}

		return Ok(Self {
			name: value.system.name.clone(),
			objects,
		});
	}
}
